#!/usr/bin/env python

# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

import symbolator
import argparse
import sys


def get_parser():
    parser = argparse.ArgumentParser(
        description="Symbolator: assess compatibility based on symbols.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    description = "actions for symbolator"
    subparsers = parser.add_subparsers(
        help="symbolator actions",
        title="actions",
        description=description,
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", help="show software version")

    # Splice a binary and find all libraries
    splice = subparsers.add_parser(
        "splice",
        help="Assess the entire space of libs for a binary, optionally splicing in a new library.",
    )

    # Jsonsplice is similar, but from input json instead
    jsonsplice = subparsers.add_parser(
        "jsonsplice",
        help="Do a splice from generate (json) output.",
    )

    for command in [splice, jsonsplice]:
        command.add_argument(
            "binary", help="Single binary to find and assess libs for.", nargs=1
        )

        command.add_argument(
            "--splice", "-s", help="Optional libraries to splice in.", action="append"
        )
        command.add_argument(
            "--dump",
            dest="dump",
            help="Dump asp to stdout instead",
            default=False,
            action="store_true",
        )

    # Stability test using smeagle output
    stability = subparsers.add_parser(
        "stability-test", help="Run a stability test using Smeagle facts."
    )

    stability.add_argument(
        "libs", help="Working and contender library, in that order", nargs=2
    )
    stability.add_argument(
        "--detail", default=False, action="store_true", help="Show detailed results."
    )

    # Compare two library elf symbols
    compare = subparsers.add_parser(
        "compare", help="Compare symbols between two libraries."
    )
    compare.add_argument(
        "libs", help="Two libraries (same but different versions) to compare", nargs=2
    )

    # Assess compatibility
    compat = subparsers.add_parser(
        "compat",
        help="Assess compatibility of a known working binary and library to a contender library",
    )
    compat.add_argument("binary", help="Main binary of interest")
    compat.add_argument(
        "libs", help="Working and contender library, in that order", nargs=2
    )
    compat.add_argument(
        "--dump",
        dest="dump",
        help="Dump asp to stdout instead",
        default=False,
        action="store_true",
    )

    # Just generate facts to the screen
    generate = subparsers.add_parser(
        "generate", help="Dump symbols as facts to the terminal."
    )
    generate.add_argument("binary", help="Main binary of interest")
    generate.add_argument(
        "--system-libs",
        dest="system_libs",
        help="Include facts for linked system libs",
        default=False,
        action="store_true",
    )
    generate.add_argument(
        "--globals",
        dest="globals_only",
        help="Only include global symbols",
        default=False,
        action="store_true",
    )

    # Either command can accept json
    for command in [generate, compat, compare, splice, jsonsplice]:
        command.add_argument(
            "--json",
            dest="json",
            help="If generating facts, dump as json instead of ASP.",
            default=False,
            action="store_true",
        )
    return parser


def run():
    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """
        version = symbolator.__version__

        print("\nSymbolator v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    # Show the version and exit
    if args.command == "version" or args.version:
        print(symbolator.__version__)
        sys.exit(0)

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser
                break

    if args.command == "compat":
        from .compat import is_compatible as main
    elif args.command == "compare":
        from .compare import compare_libs as main
    elif args.command == "generate":
        from .generate import generate as main
    elif args.command == "jsonsplice":
        from .splice import jsonsplice as main
    elif args.command == "splice":
        from .splice import splice as main
    elif args.command == "stability-test":
        from .smeagle import stability_test as main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    run()



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

"""A smeagle model will load one or more Json corpora, and do a solve.
This is adopted from https://github.com/buildsi/smeagle-demo (db)
"""

import sys
import os
import json
import jsonschema
import clingo
import pprint

from symbolator.facts import get_facts
from symbolator.asp import AspFunction, AspFunctionBuilder, Result, PyclingoDriver
from symbolator.utils import read_json

from .schema import model_schema


# Smeagle corpus schema

clingo_cffi = hasattr(clingo.Symbol, "_rep")


fn = AspFunctionBuilder()


class SmeagleClingoDriver(PyclingoDriver):
    def solve(
        self,
        setup,
        nmodels=0,
        stats=False,
        logic_programs=None,
        facts_only=False,
    ):
        """
        Run the solver for a model and some number of logic programs
        """
        # logic programs to give to the solver
        logic_programs = logic_programs or []
        if not isinstance(logic_programs, list):
            logic_programs = [logic_programs]

        # Initialize the control object for the solver
        self.control = clingo.Control()
        self.control.configuration.solve.models = nmodels
        self.control.configuration.asp.trans_ext = "all"
        self.control.configuration.asp.eq = "5"
        self.control.configuration.configuration = "tweety"
        self.control.configuration.solve.parallel_mode = "2"
        self.control.configuration.solver.opt_strategy = "usc,one"

        # set up the problem -- this generates facts and rules
        self.assumptions = []
        with self.control.backend() as backend:
            self.backend = backend
            setup.setup(self)

        # If we only want to generate facts, cut out early
        if facts_only:
            return

        # read in provided logic programs
        for logic_program in logic_programs:
            self.control.load(logic_program)

        # Grounding is the first step in the solve -- it turns our facts
        # and first-order logic rules into propositional logic.
        self.control.ground([("base", [])])

        # With a grounded program, we can run the solve.
        result = Result()
        models = []  # stable models if things go well
        cores = []  # unsatisfiable cores if they do not

        def on_model(model):
            models.append((model.cost, model.symbols(shown=True, terms=True)))

        # Won't work after this, need to write files
        solve_kwargs = {
            "assumptions": self.assumptions,
            "on_model": on_model,
            "on_core": cores.append,
        }
        if clingo_cffi:
            solve_kwargs["on_unsat"] = cores.append
        solve_result = self.control.solve(**solve_kwargs)

        # once done, construct the solve result
        result.satisfiable = solve_result.satisfiable

        def stringify(x):
            if clingo_cffi:
                # Clingo w/ CFFI will throw an exception on failure
                try:
                    return x.string
                except RuntimeError:
                    return str(x)
            else:
                return x.string or str(x)

        if result.satisfiable:
            min_cost, best_model = min(models)
            result.answers = {}
            for sym in best_model:
                if sym.name not in result.answers:
                    result.answers[sym.name] = []
                result.answers[sym.name].append([stringify(a) for a in sym.arguments])

        elif cores:
            symbols = dict((a.literal, a.symbol) for a in self.control.symbolic_atoms)
            for core in cores:
                core_symbols = []
                for atom in core:
                    sym = symbols[atom]
                    core_symbols.append(sym)
                result.cores.append(core_symbols)

        if stats:
            print("Statistics:")
            pprint.pprint(self.control.statistics)
        return result


class SolverBase:
    """
    Common base functions for some kind of solver.

    For stability, compatibility, or just fact generation.
    """

    def setup(self, driver):
        """
        Setup to prepare for the solve.
        """
        self.gen = driver

    def print(self, data, title):
        """
        Print a result to the terminal
        """
        if data:
            print("\n" + title)
            print("---------------")
            for entry in data:
                print(" " + " ".join(entry))


class StabilitySolver(SolverBase):
    """
    Class to orchestrate a Stability Solver.
    """

    def __init__(self, lib1, lib2):
        """
        Create a driver to run a compatibility model test for two libraries.
        """
        # The driver will generate facts rules to generate an ASP program.
        self.driver = SmeagleClingoDriver()
        self.setup = StabilitySolverSetup(lib1, lib2)

    def solve(self, logic_programs, detail=False, return_result=False):
        """
        Run the solve
        """
        result = self.driver.solve(self.setup, logic_programs=logic_programs)
        if return_result:
            return result
        missing_imports = result.answers.get("missing_imports", [])
        missing_exports = result.answers.get("missing_exports", [])
        if missing_imports or missing_exports:
            print(
                "Libraries are not stable: %s missing exports, %s missing_imports"
                % (len(missing_exports), len(missing_imports))
            )
            if detail:
                self.print(missing_imports, "Missing Imports")
                self.print(missing_exports, "Missing Exports")


class FactGenerator(SolverBase):
    """
    Class to orchestrate fact generation (uses FactGeneratorSetup)
    """

    def __init__(self, lib):
        """
        Create a driver to run a compatibility model test for two libraries.
        """
        # The driver will generate facts rules to generate an ASP program.
        self.driver = SmeagleClingoDriver(out=sys.stdout)
        self.setup = FactGeneratorSetup(lib)

    def solve(self):
        """
        Generate facts
        """
        return self.driver.solve(self.setup, facts_only=True)


class GeneratorBase:
    """
    The GeneratorBase is the base for any kind of Setup (fact generator or solve)
    Base functions to set up an ABI Stability and Compatability Solver.
    """

    def add_library(self, lib, identifier=None):
        """
        Given a loaded Smeagle Model, generate facts for it.
        """
        self.gen.h2("Library: %s" % lib.name)

        # Generate a fact for each location
        for loc in lib.data.get("locations", []):

            # Functions
            self.generate_function(lib, loc.get("function"), identifier)

    def generate_function(self, lib, func, identifier=None):
        """
        Generate facts for a function
        """
        if not func:
            return

        libname = os.path.basename(lib.data["library"])
        seen = set()

        for param in func.get("parameters", []):

            # These values assume no underlying type (defaults)
            param_name = param.get("name")

            # Allow to not know the location or direction
            loc = param.get("location", "unknown")
            direction = param.get("direction", "unknown")

            # We can't really do anything with an unknown name
            if not param_name:
                print("Warning, parameter without name! %s" % param)
                continue

            param_type = param.get(
                "class", "Unknown"
            )  # param['type'] is compiler specific

            # If the param has fields, continue printing until we are done
            fields = param.get("fields", [])

            # If we have an underlying type, use name, type, from there
            if "underlying_type" in param:
                param_name = param["underlying_type"].get("name") or param_name

                # Use these fields (unless they aren't defined)
                param_type = param["underlying_type"].get("class") or param_type

                # If the param has fields, continue printing until we are done
                fields = param["underlying_type"].get("fields", []) or fields

            # We are skipping locations for now - not correct
            # Location and direction are always with the original parameter
            self.gen.fact(
                fn.abi_typelocation(
                    libname,
                    func["name"],
                    param_name,
                    param_type,
                    loc,
                    direction,
                    param.get("indirections", "0"),
                )
            )

            # While we have fields, keep adding them as facts until no more
            while fields:
                field = fields.pop(0)
                self.gen.fact(
                    # The library, function name, direction and location are the same
                    fn.abi_typelocation(
                        libname,
                        func["name"],
                        field.get("name", ""),
                        field.get("class", ""),
                        loc,
                        direction,
                        field.get("indirections", "0"),
                    )
                )
                # Fields can have nested fields
                fields += field.get("fields", [])

            # If no identifier, skip the last step
            if not identifier:
                continue

            # This is only needed for the stability model to identify membership
            # of a particular function symbol, etc. with a library set (e.g., a or b)
            # Symbol, Type, Register, Direction, Pointer Indirections
            args = [
                func["name"],
                param_type,
                loc,
                direction,
                param.get("indirections", "0"),
            ]
            fact = AspFunction("is_%s" % identifier, args=args)
            if fact not in seen:
                self.gen.fact(fact)
                seen.add(fact)


class StabilitySolverSetup(GeneratorBase):
    """
    Class to set up and run an ABI Stability and Compatability Solver.
    """

    def __init__(self, lib1, lib2):
        self.lib1 = lib1
        self.lib2 = lib2

    def setup(self, driver):
        """
        Setup to prepare for the solve.

        This function overrides the base setup, which will generate facts only
        for one function.
        """
        self.gen = driver
        self.gen.h1("Library Facts")
        self.add_library(self.lib1, "a")
        self.add_library(self.lib2, "b")


class FactGeneratorSetup(GeneratorBase):
    """
    Class to accept one library and generate facts.
    """

    def __init__(self, lib):
        self.lib = lib

    def setup(self, driver):
        """
        Setup to prepare for the solve.

        This base function provides fact generation for one library.
        """
        self.gen = driver
        self.gen.h1("Library Facts")
        self.add_library(self.lib)


class Model:
    def __init__(self, name, data):
        self.name = name
        if not isinstance(data, dict):
            data = json.loads(data)
        self.data = data

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)


class SmeagleRunner:
    def __init__(self):
        """
        Load in Smeagle output files, write to database, and run solver.
        """
        self.stability_lp = get_facts("stability.lp")
        self.records = {}

    def generate_facts(self):
        """
        Generate facts for one entry.
        """
        facts = []
        for name, data in self.records.items():
            setup = FactGenerator(name)
            facts.append(setup.solve())
        return facts

    def stability_test(self, detail=False, out=None, return_result=False):
        """
        Run the stability test for two entries.
        """
        # We must have the stability program!
        if not os.path.exists(self.stability_lp):
            sys.exit("Logic program %s does not exist!" % self.stability_lp)

        # Assumes basename of libs are in database (hashes and package names)
        if len(self.records) != 2:
            sys.exit("Two libraries are required to be loaded for a stability test.")

        setup = StabilitySolver(*list(self.records.values()))
        return setup.solve(
            logic_programs=self.stability_lp, detail=detail, return_result=return_result
        )

    def load(self, path):
        """
        Load results (json) files into Smeagle database
        """
        if not os.path.exists(path):
            sys.exit("%s does not exist." % path)
        self.load_data(path)

    def load_data(self, path):
        """
        Load a json result into the sqlite database
        """
        data = read_json(path)
        name = os.path.basename(path)

        # We can only include valid models
        try:
            jsonschema.validate(data, schema=model_schema)
            self.add(name, data)
        except:
            import IPython

            IPython.embed()

    def add(self, name, data):
        """
        This does a create, only if it does not exist
        """
        if name not in self.records:
            self.records[name] = Model(name, data)



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

from symbolator.corpus import Corpus, JsonCorpusLoader
from symbolator.asp import PyclingoDriver, ABIGlobalSolverSetup
from symbolator.facts import get_facts
import json
import os
import sys


def splice(args, parser, extra, subparser):
    """

    Arguments:
        binary (str): path to a binary to assess for compataibility
        splice ([str]): list of libs to splice in.
    """
    driver = PyclingoDriver()
    if args.dump and not args.json:
        driver.out = sys.stdout

    args.splice = args.splice or []

    # Prepare splices
    lookup = {}
    for sp in args.splice:
        if "=" in sp:
            src, dest = sp.split("=")
        else:
            src, dest = sp, sp

        # The dst (the library to splice) must exist
        if not os.path.exists(dest):
            sys.exit("Splice target %s does not exist." % dest)
        lookup[src] = dest

    # We are required to have three existing libraries
    paths = args.binary + list(lookup.values())

    # Whether doing a splice or regular, we are looking for missing symbols
    logic_program = "missing_symbols.lp"

    for path in paths:
        if not os.path.exists(path):
            sys.exit("%s does not exist." % path)

    if len(paths) == 0:
        sys.exit("You must provide: <binary> [splice1]...[spliceN]")

    # No pretty printing if we are exporting json
    if not args.json:
        print("% " + "binary : %s" % args.binary[0])
        for src, dest in lookup.items():
            print("% " + "splice : %s->%s" % (src, dest))

    # Spliced libraries will be added as corpora here
    corpora = []
    for path in paths:
        corpora.append(Corpus(path))

    setup = ABIGlobalSolverSetup()

    # The order should be binary | working library | contender library
    result = driver.solve(
        setup,
        corpora,
        dump=args.dump,
        logic_programs=get_facts(logic_program),
        facts_only=False,
        splices=lookup,
    )

    missing_symbols = result.answers.get("missing_symbols")

    if missing_symbols:
        if not args.json:
            print("\nMissing Symbols:")

        # Generate lookup by library
        missing = {}
        for item in missing_symbols:
            if item[0] not in missing:
                missing[item[0]] = set()
            missing[item[0]].add(item[1])

        # Needs to be list to dump
        for lib, symbols in missing.items():
            missing[lib] = list(symbols)

        if not args.json:
            for lib, symbols in missing.items():
                print("\n=> %s" % lib)
                for symbol in symbols:
                    print("   %s" % symbol)
        else:
            print(json.dumps(missing, indent=4))
    else:
        if args.json:
            print("{}")
        else:
            print("\nThere are no missing symbols.")


def jsonsplice(args, parser, extra, subparser):
    driver = PyclingoDriver()
    if args.dump and not args.json:
        driver.out = sys.stdout

    args.splice = args.splice or []

    # Prepare splices
    lookup = {}
    for sp in args.splice:
        if "=" in sp:
            src, dest = sp.split("=")
        else:
            src, dest = sp, sp

        # The dst (the library to splice) must exist
        if not os.path.exists(dest):
            sys.exit("Splice target %s does not exist." % dest)
        lookup[src] = dest

    # We are required to have three existing libraries
    paths = args.binary + list(lookup.values())

    # Whether doing a splice or regular, we are looking for missing symbols
    logic_program = "missing_symbols.lp"

    for path in paths:
        if not os.path.exists(path):
            sys.exit("%s does not exist." % path)

    if len(paths) == 0:
        sys.exit("You must provide: <binary> [splice1]...[spliceN]")

    # No pretty printing if we are exporting json
    if not args.json:
        print("% " + "binary : %s" % args.binary[0])
        for src, dest in lookup.items():
            print("% " + "splice : %s->%s" % (src, dest))

    # Spliced libraries will be added as corpora here
    loader = JsonCorpusLoader()
    loader.load(args.binary[0])
    corpora = loader.get_lookup()

    # Now load the splices separately, and select what we need
    for lib, jsonfile in lookup.items():
        splice_loader = JsonCorpusLoader()
        splice_loader.load(jsonfile)
        splices = splice_loader.get_lookup()

        # If we have the library in corpora, delete it, add spliced libraries
        if lib in corpora:
            del corpora[lib]
            for newlib, newcorp in splices.items():
                corpora[newlib] = newcorp

    setup = ABIGlobalSolverSetup()

    # The order should be binary | working library | contender library
    result = driver.solve(
        setup,
        list(corpora.values()),
        dump=args.dump,
        logic_programs=get_facts(logic_program),
        facts_only=False,
        # Loading from json already includes system libs
        system_libs=False,
    )

    missing_symbols = result.answers.get("missing_symbols")

    if missing_symbols:
        if not args.json:
            print("\nMissing Symbols:")

        # Generate lookup by library
        missing = {}
        for item in missing_symbols:
            if item[0] not in missing:
                missing[item[0]] = set()
            missing[item[0]].add(item[1])

        # Needs to be list to dump
        for lib, symbols in missing.items():
            missing[lib] = list(symbols)

        if not args.json:
            for lib, symbols in missing.items():
                print("\n=> %s" % lib)
                for symbol in symbols:
                    print("   %s" % symbol)
        else:
            print(json.dumps(missing, indent=4))
    else:
        if args.json:
            print("{}")
        else:
            print("\nThere are no missing symbols.")



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

from symbolator.corpus import Corpus
from symbolator.asp import PyclingoDriver, ABICompatSolverSetup
from symbolator.facts import get_facts
import json
import os
import sys


def is_compatible(args, parser, extra, subparser):
    """
    Given three libraries (we call one a main binary and the other a library
    that works, and the second one that we want to link with it), determine
    if the second library is compatible with the binary. This wrapper function,
    in requiring the three named arguments for binary, libraryA and libraryB,
    enforces that all three are provided (and in the correct order).
    If the functions for the solver are used outside this context, make sure
    that you provide them in the correct order.

    Arguments:
        binary (str): path to a binary to assess for compataibility
        libraryA (str): path to a library that is known to work
        libraryB (str): a second library to assess for compatability.
        dump (tuple): what to dump
        models (int): number of models to search (default: 0)
    """
    driver = PyclingoDriver()
    if args.dump:
        driver.out = sys.stdout

    # We are required to have three existing libraries
    paths = [args.binary] + args.libs
    for path in paths:
        if not os.path.exists(path):
            sys.exit("%s does not exist." % path)

    if len(paths) != 3:
        sys.exit("You must provide: <binary> <working-lib> <contender-lib>")

    # No pretty printing if we are exporting json
    if not args.json:
        print("% " + "binary           : %s" % args.binary)
        print("% " + "working library  : %s" % args.libs[0])
        print("% " + "contender library: %s" % args.libs[1])

    corpora = []
    for path in paths:
        corpora.append(Corpus(path))
    setup = ABICompatSolverSetup()

    # The order should be binary | working library | contender library
    result = driver.solve(
        setup,
        corpora,
        dump=args.dump or not args.json,
        logic_programs=get_facts("is_compatible.lp"),
        facts_only=args.dump,
    )

    # Don't print any more if we are just dumping
    if args.dump:
        return

    if args.json:

        missing_symbols = 0
        if (
            "count_missing_symbols" in result.answers
            and result.answers["count_missing_symbols"]
        ):
            missing_symbols = result.answers["count_missing_symbols"][0]
        data = {
            "binary": corpora[0].path,
            "library_working": corpora[1].path,
            "library_contender": corpora[2].path,
            "missing_symbols": result.answers.get("missing_symbols", []),
            "count_missing_symbols": missing_symbols,
        }
        print(json.dumps(data, indent=4))
    else:
        print("Missing Symbol Count: %s" % result.answers["count_missing_symbols"][0])
        print("Missing Symbols:\n%s" % result.answers["missing_symbols"])



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

import os
import sys
import time
import types

import symbolator.utils as utils
from .corpus import Corpus

# Since we parse the die's directly, we use these pyelftools supporting functions.
from elftools.common.py3compat import bytes2str
from elftools.dwarf.descriptions import describe_attr_value
from elftools.elf.descriptions import (
    describe_symbol_type,
    describe_symbol_bind,
    describe_symbol_visibility,
    describe_symbol_shndx,
)

from six import string_types

# An arbitrary version for this asp.py (libabigail has one, so we are copying)
__version__ = "1.0.0"

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

try:
    import clingo

    # There may be a better way to detect this
    clingo_cffi = hasattr(clingo.Symbol, "_rep")
except ImportError:
    sys.exit("clingo for Python is required.")


if sys.version_info >= (3, 3):
    from collections.abc import Sequence  # novm
else:
    from collections import Sequence


class Timer(object):
    """Simple timer for timing phases of a solve"""

    def __init__(self):
        self.start = time.time()
        self.last = self.start
        self.phases = {}

    def phase(self, name):
        last = self.last
        now = time.time()
        self.phases[name] = now - last
        self.last = now

    def write(self, out=sys.stdout):
        now = time.time()
        out.write("Time:\n")
        for phase, t in self.phases.items():
            out.write("    %-15s%.4f\n" % (phase + ":", t))
        out.write("Total: %.4f\n" % (now - self.start))


def issequence(obj):
    if isinstance(obj, string_types):
        return False
    return isinstance(obj, (Sequence, types.GeneratorType))


def listify(args):
    if len(args) == 1 and issequence(args[0]):
        return list(args[0])
    return list(args)


class AspObject:
    """Object representing a piece of ASP code."""


def _id(thing):
    """Quote string if needed for it to be a valid identifier."""
    if isinstance(thing, AspObject):
        return thing
    elif isinstance(thing, bool):
        return '"%s"' % str(thing)
    elif isinstance(thing, int):
        return str(thing)
    else:
        return '"%s"' % str(thing)


class AspFunction(AspObject):
    def __init__(self, name, args=None):
        self.name = name
        self.args = [] if args is None else args

    def __call__(self, *args):
        return AspFunction(self.name, args)

    def symbol(self, positive=True):
        def argify(arg):
            if isinstance(arg, bool):
                return clingo.String(str(arg))
            elif isinstance(arg, int):
                return clingo.Number(arg)
            else:
                return clingo.String(str(arg))

        return clingo.Function(
            self.name, [argify(arg) for arg in self.args], positive=positive
        )

    def __getitem___(self, *args):
        self.args[:] = args
        return self

    def __str__(self):
        return "%s(%s)" % (self.name, ", ".join(str(_id(arg)) for arg in self.args))

    def __repr__(self):
        return str(self)


class AspFunctionBuilder(object):
    def __getattr__(self, name):
        return AspFunction(name)


fn = AspFunctionBuilder()


class Result(object):
    """
    Result of an ASP solve.
    """

    def __init__(self, asp=None):
        self.asp = asp
        self.satisfiable = None
        self.optimal = None
        self.warnings = None
        self.nmodels = 0

        # specs ordered by optimization level
        self.answers = []
        self.cores = []

    def print_cores(self):
        for core in self.cores:
            print(
                "The following constraints are unsatisfiable:",
                *sorted(str(symbol) for symbol in core)
            )


class PyclingoDriver:
    def __init__(self, cores=True, out=None):
        """Driver for the Python clingo interface.

        Arguments:PyclingoDriver
            cores (bool): whether to generate unsatisfiable cores for better
                error reporting.
            out (file-like): optional stream to write a text-based ASP program
                for debugging or verification.
        """
        global clingo
        if out:
            self.out = out
        else:
            self.devnull()
        self.cores = cores
        self.assumptions = []

    def devnull(self):
        self.f = open(os.devnull, "w")
        self.out = self.f

    def __exit__(self):
        self.f.close()

    def title(self, name, char):
        self.out.write("\n")
        self.out.write("%" + (char * 76))
        self.out.write("\n")
        self.out.write("%% %s\n" % name)
        self.out.write("%" + (char * 76))
        self.out.write("\n")

    def h1(self, name):
        self.title(name, "=")

    def h2(self, name):
        self.title(name, "-")

    def newline(self):
        self.out.write("\n")

    def fact(self, head):
        """ASP fact (a rule without a body)."""
        symbol = head.symbol() if hasattr(head, "symbol") else head

        self.out.write("%s.\n" % str(symbol))

        atom = self.backend.add_atom(symbol)
        self.backend.add_rule([atom], [], choice=self.cores)
        if self.cores:
            self.assumptions.append(atom)

    def solve(
        self,
        solver_setup,
        corpora,
        dump=None,
        nmodels=0,
        timers=False,
        stats=False,
        logic_programs=None,
        facts_only=False,
        # Only relevant for single library symbol dumps
        system_libs=False,
        is_single=False,
        splices=None,
    ):
        """Given three corpora, generate facts for a solver.

        The order is important:

         [binary, libraryA, libraryB]:
           binary: should be a binary that uses libraryA
           libraryA: should be a known library to work with the binary
           libraryB: should be a second library to test if it will work.

        In the future ideally we would not want to require this working library,
        but for now we are trying to emulate what libabigail does. The first
        working binary serves as a base to subset the symbols to a known set
        that are needed. We could possibly remove it if we can load all symbols
        provided by other needed files, and then eliminate them from the set.
        """
        # logic programs to give to the solver
        logic_programs = logic_programs or []
        if not isinstance(logic_programs, list):
            logic_programs = [logic_programs]

        timer = Timer()
        self.control = clingo.Control()

        # Splices get handed to the solver setup
        splices = splices or {}

        # set up the problem -- this generates facts and rules
        with self.control.backend() as backend:
            self.backend = backend

            # one corpus we can only generate facts
            if len(corpora) == 1 and is_single:
                solver_setup.single_setup(
                    self, corpora[0], system_libs=system_libs, splices=splices
                )
            else:
                solver_setup.compat_setup(
                    self, corpora, splices=splices, system_libs=system_libs
                )
        timer.phase("setup")

        # If we only want to generate facts, cut out early
        if facts_only:
            return

        for logic_program in logic_programs:
            self.control.load(logic_program)
        timer.phase("load")
        self.control.ground([("base", [])])

        # With a grounded program, we can run the solve.
        result = Result()
        models = []  # stable models if things go well
        cores = []  # unsatisfiable cores if they do not

        def on_model(model):
            models.append((model.cost, model.symbols(shown=True, terms=True)))

        solve_kwargs = {
            "assumptions": self.assumptions,
            "on_model": on_model,
            "on_core": cores.append,
        }

        if clingo_cffi:
            solve_kwargs["on_unsat"] = cores.append

        # Get the result object
        solve_result = self.control.solve(**solve_kwargs)
        timer.phase("solve")

        # once done, construct the solve result
        result.satisfiable = solve_result.satisfiable

        def stringify(x):
            if clingo_cffi:
                # Clingo w/ CFFI will throw an exception on failure
                try:
                    return x.string
                except RuntimeError:
                    return str(x)
            else:
                return x.string or str(x)

        if result.satisfiable:
            min_cost, best_model = min(models)
            result.answers = {}
            for sym in best_model:
                if sym.name in result.answers:
                    result.answers[sym.name].append(
                        [stringify(a) for a in sym.arguments]
                    )
                else:
                    result.answers[sym.name] = [[stringify(a) for a in sym.arguments]]

        elif cores:
            symbols = dict((a.literal, a.symbol) for a in self.control.symbolic_atoms)
            for core in cores:
                core_symbols = []
                for atom in core:
                    sym = symbols[atom]
                    if sym.name == "missing_symbol":
                        sym = sym.arguments[0].string
                    core_symbols.append(sym)
                result.cores.append(core_symbols)

        if timers:
            timer.write()
            print()

        # statistics
        # pprint.pprint(self.control.statistics)

        return result


class ABISolverBase:
    """
    Base class with shared functions
    """

    def generate_elf_symbols(self, corpora, prefix=""):
        """For each corpus, write out elf symbols as facts. Note that we are
        trying a more detailed approach with facts/atoms being named (e.g.,
        symbol_type instead of symbol_attr). We could also try a simpler
        approach:

        symbol_type("_ZN11MathLibrary10Arithmetic8MultiplyEdd", "STT_FUNC").
        symbol_binding("_ZN11MathLibrary10Arithmetic8MultiplyEdd", "STB_FUNC").
        symbol_attr("_ZN11MathLibrary10Arithmetic8MultiplyEdd", "STV_default").
        """
        # If we have a prefix, add a spacer
        prefix = "%s_" % prefix if prefix else ""

        for corpus in corpora:
            self.gen.h2("Corpus symbols: %s" % corpus.path)

            for symbol, meta in corpus.symbols.items():

                # It begins with a NULL symbol, not sure it's useful
                if not symbol:
                    continue

                # If we have @@ in the symbol, it's usually the compiler (remove)
                if "@@" in symbol:
                    symbol = symbol.split("@@")[0]

                self.gen.fact(AspFunction(prefix + "symbol", args=[symbol]))
                self.gen.fact(
                    AspFunction(
                        prefix + "symbol_type", args=[corpus.path, symbol, meta["type"]]
                    )
                )
                self.gen.fact(
                    AspFunction(
                        prefix + "symbol_version",
                        args=[corpus.path, symbol, meta["version_info"]],
                    )
                )
                self.gen.fact(
                    AspFunction(
                        prefix + "symbol_binding",
                        args=[corpus.path, symbol, meta["binding"]],
                    )
                )
                self.gen.fact(
                    AspFunction(
                        prefix + "symbol_visibility",
                        args=[corpus.path, symbol, meta["visibility"]],
                    )
                )
                self.gen.fact(
                    AspFunction(
                        prefix + "symbol_definition",
                        args=[corpus.path, symbol, meta["defined"]],
                    )
                )

                # Might be redundant
                has = "has_%s" % prefix if prefix else "has_"
                self.gen.fact(AspFunction(has + "symbol", args=[corpus.path, symbol]))
                self.gen.fact(fn.has_symbol(corpus.path, symbol))

    def generate_needed(self, corpora):
        """
        Given a list of corpora, add needed libraries from dynamic tags.
        """
        for corpus in corpora:
            for needed in corpus.dynamic_tags.get("needed", []):
                self.gen.fact(fn.corpus_needs_library(corpus.path, needed))

    def generate_corpus_metadata(self, corpora, prefix=""):
        """Given a list of corpora, create a fact for each one. If we need them,
        we can add elfheaders here.
        """
        prefix = "%s_" % prefix if prefix else ""

        # Use the corpus path as a unique id (ok if binaries exist)
        # This would need to be changed if we don't have the binary handy
        for corpus in corpora:
            hdr = corpus.elfheader

            self.gen.h2("Corpus facts: %s" % corpus.path)

            self.gen.fact(fn.corpus(corpus.path))
            self.gen.fact(AspFunction(prefix + "corpus", args=[corpus.path]))
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_name",
                    args=[corpus.path, os.path.basename(corpus.path)],
                )
            )

            # e_ident is ELF identification
            # https://docs.oracle.com/cd/E19683-01/816-1386/chapter6-35342/index.html
            # Note that we could update these to just be corpus_attr, but I'm
            # starting with testing a more detailed approach for now.

            # If the corpus has a soname:
            if corpus.soname:
                self.gen.fact(
                    AspFunction(
                        prefix + "corpus_soname", args=[corpus.path, corpus.soname]
                    )
                )

            # File class (also at elffile.elfclass or corpus.elfclass
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_elf_class",
                    args=[corpus.path, hdr["e_ident"]["EI_CLASS"]],
                )
            )

            # Data encoding
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_data_encoding",
                    args=[corpus.path, hdr["e_ident"]["EI_DATA"]],
                )
            )

            # File version
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_file_version",
                    args=[corpus.path, hdr["e_ident"]["EI_VERSION"]],
                )
            )

            # Operating system / ABI Information
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_elf_osabi",
                    args=[corpus.path, hdr["e_ident"]["EI_OSABI"]],
                )
            )

            # Abi Version
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_abiversion",
                    args=[corpus.path, hdr["e_ident"]["EI_ABIVERSION"]],
                )
            )

            # e_type is the object file type
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_elf_type", args=[corpus.path, hdr["e_type"]]
                )
            )

            # e_machine is the required architecture for the file
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_elf_machine", args=[corpus.path, hdr["e_machine"]]
                )
            )

            # object file version
            self.gen.fact(
                AspFunction(
                    prefix + "corpus_elf_version", args=[corpus.path, hdr["e_version"]]
                )
            )

    def get_system_corpora(self, corpora):
        """
        Get a list of corpora for system corpora. If we are doing splicing,
        honor the splice instead.
        """
        ldd = utils.which("ldd").get("message")
        if not ldd:
            print("Cannot find ldd to detect system libraries, skipping.")
            return

        # Add the pwd and LD_LIBRARY_PATH to path
        ld_libs = os.environ.get("LD_LIBRARY_PATH")
        here = os.getcwd()
        path = os.environ.get("PATH") + ":%s" % here
        if ld_libs:
            path = "%s:%s" % (path, ld_libs)
            ld_libs = "%s:%s" % (ld_libs, here)
        else:
            ld_libs = here

        os.environ["PATH"] = path
        os.putenv("PATH", path)
        os.putenv("LD_LIBRARY_PATH", ld_libs)

        # Ensure we don't add a library twice
        seen = set([x.path for x in corpora])
        syscorpora = []

        for corpus in corpora:

            # Try adding rpaths
            rpaths = corpus.dynamic_tags.get("rpaths")
            if rpaths:
                os.putenv("LD_LIBRARY_PATH", "%s:%s" % (ld_libs, ":".join(rpaths)))

            output = utils.run_command([ldd, corpus.path]).get("message", "")

            for line in output.split("\n"):

                # If we cannot find a lib, still might be able to splice
                path = None
                lib = None
                if "not found" in line:
                    lib = line.split("=>")[0].strip()
                    if lib in corpus.needed:
                        path = lib

                # Case 2: a library was found
                elif "=>" in line:
                    lib, path = line.split("=>")
                    lib = lib.strip()
                    path = path.strip().split(" ")[0]

                if path:
                    # Don't add redundant entries
                    if path in seen or os.path.basename(path) in seen:
                        continue

                    # If it's a splice entry, make sure we've added corpus
                    if path in self.splices:
                        path = self.splices[path]
                    elif os.path.basename(path) in self.splices:
                        path = self.splices[os.path.basename(path)]

                    if os.path.exists(path):
                        syscorpora.append(Corpus(path, name=lib))
                    elif not os.path.exists(path) and lib in corpus.needed:
                        print(
                            "Warning: %s is needed, but not found on system path."
                            % path
                        )
                    seen.add(path)

        return syscorpora

    def single_setup(self, driver, corpus, system_libs=False, **kwargs):
        """
        Generate facts for a single library.
        """
        assert corpus.exists()
        self.splices = kwargs.get("splices", {})

        # driver is used by all the functions below to add facts and
        # rules to generate an ASP program.
        self.gen = driver

        self.gen.h1("Corpus Facts")

        # Generate high level corpus metadata facts (e.g., header)
        self.generate_corpus_metadata([corpus])

        # Dynamic libraries that are needed
        self.generate_needed([corpus])

        # generate all elf symbols (might be able to make this smaller set)
        self.generate_elf_symbols([corpus])

        # Add system corpora to elf symbols
        if system_libs:
            self.generate_elf_symbols(self.get_system_corpora([corpus]))

    def get_metadata(self, corpus):
        """
        Dump corpus metadata to json
        """
        hdr = corpus.elfheader

        # Corpus metadata
        return {
            "path": corpus.path,
            "corpus_name": os.path.basename(corpus.path),
            "corpus_soname": corpus.soname,
            "corpus_elf_class": hdr["e_ident"]["EI_CLASS"],
            "corpus_data_encoding": hdr["e_ident"]["EI_DATA"],
            "corpus_file_version": hdr["e_ident"]["EI_VERSION"],
            "corpus_elf_osabi": hdr["e_ident"]["EI_OSABI"],
            "corpus_abiversion": hdr["e_ident"]["EI_ABIVERSION"],
            "corpus_elf_type": hdr["e_type"],
            "corpus_elf_machine": hdr["e_machine"],
            "corpus_elf_version": hdr["e_version"],
        }

    def get_json(self, corpus, system_libs=False, corpora=None, **kwargs):
        """
        Get json symbols and metadata instead.
        """
        # We can call this recursively and append
        corpora = corpora or []
        assert corpus.exists()
        self.splices = kwargs.get("splices", {})
        globals_only = kwargs.get("globals_only", False)

        data = {"corpus": {}}
        data["corpus"]["metadata"] = self.get_metadata(corpus)

        # This needs to be json serializable
        hdr = corpus.elfheader
        hdr["e_ident"] = dict(hdr.get("e_ident", {}))

        # Needed and dynamic tags
        data["corpus"]["needed"] = corpus.dynamic_tags.get("needed", [])
        data["corpus"]["dynamic_tags"] = corpus.dynamic_tags
        data["corpus"]["header"] = hdr

        # Elf symbols
        symbols = corpus.symbols
        if globals_only:
            updated = {}
            for name, symbol in symbols.items():
                if symbol["binding"] != "GLOBAL":
                    continue
                updated[name] = symbol
            symbols = updated

        data["corpus"]["symbols"] = symbols
        corpora.append(data)

        # Add system corpora to elf symbols
        if system_libs:
            for lib in self.get_system_corpora([corpus]):

                # Do not recursively add system libs - we can only care about top level corpus
                corpora += self.get_json(
                    lib, system_libs=False, corpora=corpora, **kwargs
                )

        # One final filter to only add unique paths
        uniques = []
        seen = set()
        for corpus in corpora:
            path = corpus["corpus"]["metadata"]["path"]
            if path in seen:
                continue
            uniques.append(corpus)
            seen.add(path)
        return uniques


class ABICompatSolverSetup(ABISolverBase):
    """
    Class to set up and run a compatibility check for a binary and two libs.
    """

    def compat_setup(self, driver, corpora, **kwargs):
        """
        Generate an ASP program with relevant constraints for a binary
        and a library, for which we have been provided their corpora.

        This calls methods on the solve driver to set up the problem with
        facts and rules from both corpora, and rules that determine ABI
        compatibility.

        Arguments:
            corpora: [corpusA, corpusB, corpusC]
            corpusA (corpus.Corpus): the first corpus for the binary
            corpusB (corpus.Corpus): the corpus for the library that works
            corpusB (corpus.Corpus): the corpus for the library that we test

        """
        # preliminary checks
        for corpus in corpora:
            assert corpus.exists()

        binary, working, contender = corpora
        corpora = [binary, contender]
        self.splices = kwargs.get("splices", {})
        system_libs = kwargs.get("system_libs") or True

        # driver is used by all the functions below to add facts and
        # rules to generate an ASP program.
        self.gen = driver

        self.gen.h1("Corpus Facts")

        self.gen.fact(fn.is_main(binary.path))
        self.gen.fact(fn.is_library(contender.path))
        self.gen.fact(fn.is_needed(working.path))

        # Generate high level corpus metadata facts (e.g., header)
        self.generate_corpus_metadata(corpora)
        self.generate_corpus_metadata([working], prefix="needed")

        # Dynamic libraries that are needed
        self.generate_needed(corpora)

        # generate all elf symbols (might be able to make this smaller set)
        self.generate_elf_symbols(corpora)

        # Generate the same for the known working library, but with a prefix
        self.generate_elf_symbols([working], prefix="needed")

        # Add system corpora to elf symbols
        if system_libs:
            self.generate_elf_symbols(self.get_system_corpora(corpora))


class ABIGlobalSolverSetup(ABISolverBase):
    """
    Class to set up and run a compatibility check for a binary and some number of libs.
    """

    def compat_setup(self, driver, corpora, **kwargs):
        self.splices = kwargs.get("splices", {})
        system_libs = kwargs.get("system_libs", True)

        # driver is used by all the functions below to add facts and
        # rules to generate an ASP program.
        self.gen = driver

        self.gen.h1("Corpus Facts")

        # Generate high level corpus metadata facts (e.g., header)
        self.generate_corpus_metadata(corpora)

        # Dynamic libraries that are needed
        self.generate_needed(corpora)

        # generate all elf symbols (might be able to make this smaller set)
        self.generate_elf_symbols(corpora)

        # Add system corpora to elf symbols
        if system_libs:
            self.generate_elf_symbols(self.get_system_corpora(corpora))


class ABICompareSolverSetup(ABISolverBase):
    """
    Class to set up and run a comparison between two libraries
    """

    def compat_setup(self, driver, corpora, **kwargs):
        """
        Arguments:
            corpora: [corpusA, corpusB]
            corpusB (corpus.Corpus): the first library corpus
            corpusB (corpus.Corpus): the second library corpus
        """
        # preliminary checks
        for corpus in corpora:
            assert corpus.exists()

        libA, libB = corpora
        self.splices = kwargs.get("splices", {})
        system_libs = kwargs.get("system_libs") or True

        # driver is used by all the functions below to add facts and
        # rules to generate an ASP program.
        self.gen = driver

        self.gen.h1("Corpus Facts")

        self.gen.fact(fn.is_libA(libA.path))
        self.gen.fact(fn.is_libB(libB.path))

        # Generate high level corpus metadata facts (e.g., header)
        self.generate_corpus_metadata(corpora)

        # Dynamic libraries that are needed
        self.generate_needed(corpora)

        # generate all elf symbols (might be able to make this smaller set)
        self.generate_elf_symbols(corpora)

        # Add system corpora to elf symbols
        if system_libs:
            self.generate_elf_symbols(self.get_system_corpora(corpora))



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

# from symbolator.corpus import Corpus
# from symbolator.asp import PyclingoDriver, ABICompatSolverSetup
# from symbolator.facts import get_facts

from symbolator.smeagle import SmeagleRunner


def stability_test(args, parser, extra, subparser):
    """
    Run a stability test with Smeagle.
    """
    smeagle = SmeagleRunner()

    # Load the libraries
    for lib in args.libs:
        smeagle.load(lib)

    # Stability test between two libraries
    smeagle.stability_test(detail=args.detail)



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

from symbolator.corpus import Corpus
from symbolator.asp import PyclingoDriver, ABICompatSolverSetup
import json
import io

# Functions intended to be called by external clients


def generate(args, parser, extra, subparser):
    """
    A single function to print facts for one or more corpora.
    """
    setup = ABICompatSolverSetup()
    corpus = Corpus(args.binary)

    # Json output
    if args.json:
        result = setup.get_json(
            corpus, system_libs=args.system_libs, globals_only=args.globals_only
        )
        print(json.dumps(result, indent=4))

    # Asp output
    else:
        # Get output via StringIO
        out = io.StringIO()
        driver = PyclingoDriver(out=out)
        driver.solve(
            setup,
            [corpus],
            facts_only=True,
            system_libs=args.system_libs,
            is_single=True,
        )
        asp = out.getvalue()
        out.close()
        print(asp)



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

"""A Corpus object is used to extract Dwarf and ELF symbols from a library.
We exclude most of the Dwarf Information Entries to expose a reduced set.
Entries can be added as they are needed.
"""

from symbolator.utils import read_json
import sys
import os


class CorpusBase:
    def __init__(self, filename, name=None, uid=None, **kwargs):

        must_exist = kwargs.get("must_exist", True)
        if not os.path.exists(filename) and must_exist:
            sys.exit("%s does not exist." % filename)

        self.elfheader = {}
        self.basename = os.path.basename(filename)

        self.name = name
        self.uid = uid
        self.symbols = {}
        self.path = filename
        self.basename = None
        self.dynamic_tags = {}
        self.architecture = None
        self._soname = None
        self.kwargs = kwargs
        self.read_corpus()

    def read_corpus(self):
        raise NotImplementedError

    def __str__(self):
        return "[Corpus:%s]" % self.path

    def __repr__(self):
        return str(self)

    def exists(self):
        return self.path is not None and os.path.exists(self.path)

    @property
    def soname(self):
        return self.dynamic_tags.get("soname")

    @property
    def needed(self):
        return self.dynamic_tags.get("needed", [])

    @property
    def runpath(self):
        return self.dynamic_tags.get("runpath")

    @property
    def rpath(self):
        return self.dynamic_tags.get("rpath")


class JsonCorpusLoader:
    def __init__(self):
        self.seen = set()
        self.corpora = []

    def get_lookup(self, key="name"):
        """
        Return a corpus lookup (by key)
        """
        # Create a lookup we can easily sub a library in for
        lookup = {}
        for corpus in self.corpora:
            field = getattr(corpus, key)
            if field in lookup:
                print("Warning: %s is seen more than once in lookup." % field)
            lookup[field] = corpus
        return lookup

    def load(self, content):
        """
        Given a json dump of a corpus (and system libraries) load into corpora.
        Make sure to split symbols to not include @
        """
        # If it isn't already loaded!
        if not isinstance(content, list):
            if not os.path.exists(content):
                sys.exit("%s for loading corpora does not exist." % content)
            content = read_json(content)

        for entry in content:
            if "corpus" not in entry:
                sys.exit("corpus key missing at top level!")
            corpus = entry["corpus"]
            filename = corpus["metadata"]["path"]
            name = corpus["metadata"]["corpus_name"]

            # It's reasonable we could have a corpus seen twice
            if filename not in self.seen:
                self.corpora.append(
                    JsonCorpus(filename, name, loaded=entry["corpus"], must_exist=False)
                )
                self.seen.add(filename)


class JsonCorpus(CorpusBase):
    """
    Generate an ABI corpus from Json input
    """

    def read_corpus(self):
        """
        Read the entire json corpus
        """
        loaded = self.kwargs.get("loaded")
        if not loaded:
            sys.exit("Cannot create a json corpus without loaded content.")

        self.metadata = loaded.get("metadata", {})
        self.elfheader = loaded.get("header", {})
        self.dynamic_tags = loaded.get("dynamic_tags", {})
        self.architecture = self.metadata.get("corpus_elf_machine")
        self.elfclass = self.metadata.get("corpus_elf_class")

        # Ensure we remove @s from symbols - we don't care about compiler version
        symbols = loaded.get("symbols", {})
        self.symbols = {}
        for name, meta in symbols.items():
            name = name.split("@")[0]
            self.symbols[name] = meta



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

from symbolator.corpus import Corpus
from symbolator.asp import PyclingoDriver, ABICompareSolverSetup
from symbolator.facts import get_facts
import json
import os
import sys


def compare_libs(args, parser, extra, subparser):
    """
    Given two libraries (same but different versions) compare across symbols.
    If Smeagle output is provided, we do a more detailed comparison.

    Arguments:
        libraryA (str): path to the first library
        libraryB (str): path to the second library
        json (bool): if the input is smeagle json
    """
    driver = PyclingoDriver()
    if not args.json:
        driver.out = sys.stdout

    # We are required to have three existing libraries
    for path in args.libs:
        if not os.path.exists(path):
            sys.exit("%s does not exist." % path)

    if len(args.libs) != 2:
        sys.exit("You must provide: <first-lib> <second-lib>")

    if not args.json:
        print("% " + "first library : %s" % args.libs[0])
        print("% " + "second library: %s" % args.libs[1])

    corpora = []
    for path in args.libs:
        corpora.append(Corpus(path))
    setup = ABICompareSolverSetup()

    # The order should be binary | working library | contender library
    result = driver.solve(
        setup,
        corpora,
        dump=not args.json,
        logic_programs=get_facts("compare_libs.lp"),
        facts_only=False,
    )

    print(json.dumps(result.answers, indent=4))



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

__version__ = "0.0.18"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsoch@users.noreply.github.com"
NAME = "symbolator-python"
PACKAGE_URL = "https://github.com/buildsi/symbolator"
KEYWORDS = "abi, compatibility, symbols"
DESCRIPTION = "Assess compatibility of two binaries based on symbols."
LICENSE = "LICENSE"

################################################################################
# Global requirements

INSTALL_REQUIRES = (
    ("pyelftools", {"min_version": None}),
    ("jsonschema", {"min_version": None}),
    ("clingo", {"min_version": None}),
    ("six", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (versions that include database)

INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

# Fields schema

fields = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["name"],
        # We allow additional properties here, because we can have additional underlying types
        "additionalProperties": True,
        "properties": {
            "name": {"type": "string"},
            "class": {"type": "string"},
            "type": {"type": "string"},
            "size": {"type": ["string", "number"]},
            # TODO should we add validation of when fields is allowed?
            # This indicates that fields can be recursive
            "fields": {"$ref": "#"},
        },
    },
}

# Underlying type schema
underlying_type = {
    "type": "object",
    "additionalProperties": False,
    "properties": {
        "name": {"type": "string"},
        "class": {"type": "string"},
        "type": {"type": "string"},
        "size": {"type": ["string", "number"]},
        "fields": fields,
    },
}


# Schema for smeagle model
model_schema = {
    "$schema": "http://json-schema.org/schema#",
    "title": "build-abi-containers package schema",
    "type": "object",
    "additionalProperties": False,
    "required": ["library", "locations"],
    "properties": {
        "library": {"type": "string"},
        "locations": {
            "type": "array",
            "items": {
                "type": "object",
                "required": ["function"],
                "properties": {
                    "function": {
                        "type": "object",
                        "required": ["name"],
                        "properties": {
                            "name": {"type": "string"},
                            "parameters": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "additionalProperties": False,
                                    "properties": {
                                        "name": {"type": "string"},
                                        "indirections": {"type": ["string", "null"]},
                                        "class": {"type": "string"},
                                        "type": {"type": "string"},
                                        "size": {"type": ["string", "number"]},
                                        "location": {"type": "string"},
                                        "fields": fields,
                                        "underlying_type": underlying_type,
                                        "direction": {
                                            "type": ["string", "number"],
                                            "enum": ["import", "export", "unknown"],
                                        },
                                    },
                                },
                            },
                        },
                    }
                },
            },
        },
    },
}



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

import os

here = os.path.abspath(os.path.dirname(__file__))


def get_facts(name):
    filename = os.path.join(here, name)
    if os.path.exists(filename):
        return filename



# Copyright 2013-2021 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)

"""A Corpus object is used to extract Dwarf and ELF symbols from a library.
We exclude most of the Dwarf Information Entries to expose a reduced set.
Entries can be added as they are needed.
"""

import sys
from .base import CorpusBase


class ElftoolsWrapper(object):
    """Create a wrapper to elftools that we can share to expose subfunctions."""

    def __init__(self):
        import elftools
        from elftools.elf import (
            dynamic,
            descriptions,
            sections,
            gnuversions,
            elffile,
            constants,
        )
        from elftools.dwarf import descriptions as dwarf
        from elftools.dwarf import locationlists as locationlists
        from elftools.common import py3compat as py3compat
        from elftools.common import exceptions as exceptions

        self.dynamic = dynamic
        self.descriptions = descriptions
        self.sections = sections
        self.gnuversions = gnuversions
        self.elffile = elffile
        self.constants = constants
        self.dwarf = dwarf
        self.locationlists = locationlists
        self.py3compat = py3compat
        self.exceptions = exceptions


et = ElftoolsWrapper()


class CorpusReader(et.elffile.ELFFile):  # noqa
    """
    A CorpusReader wraps an elffile.

    This allows us to easily open/close and keep the stream open while we are
    interacting with content. We close the file handle on any exit.
    """

    def __init__(self, filename):
        self.fd = open(filename, "rb")
        self.filename = filename
        try:
            self.elffile = et.elffile.ELFFile(self.fd)
        except Exception:
            sys.exit("%s is not an ELF file." % filename)

        # Cannot continue without dwarf info
        if not self.elffile.has_dwarf_info():
            sys.exit("%s is missing DWARF info." % self.filename)
        self.get_version_lookup()
        self.get_shndx_sections()

    def __str__(self):
        return "[CorpusReader:%s]" % self.filename

    def __repr__(self):
        return str(self)

    @property
    def header(self):
        return dict(self.elffile.header)

    def __exit__(self):
        self.fd.close()

    def get_architecture(self):
        return self.elffile.header.get("e_machine")

    def get_elf_class(self):
        return self.elffile.elfclass

    def get_version_lookup(self):
        """
        Get versioning used (GNU or Solaris)
        https://github.com/eliben/pyelftools/blob/master/scripts/readelf.py#L915
        """
        lookup = dict()
        types = {
            et.gnuversions.GNUVerSymSection: "versym",
            et.gnuversions.GNUVerDefSection: "verdef",
            et.gnuversions.GNUVerNeedSection: "verneed",
            et.dynamic.DynamicSection: "type",
        }

        for section in self.elffile.iter_sections():
            if type(section) in types:
                identifier = types[type(section)]
                if identifier == "type":
                    for tag in section.iter_tags():
                        if tag["d_tag"] == "DT_VERSYM":
                            lookup["type"] = "GNU"
                else:
                    lookup[identifier] = section

        # If we don't have a type but we have verneed or verdef, it's solaris
        if not lookup.get("type") and (lookup.get("verneed") or lookup.get("verdef")):
            lookup["type"] = "Solaris"
        self._versions = lookup

    def get_shndx_sections(self):
        """
        Get section lookup.

        I think this referes to section index/indices. We want a mapping
        from a symbol table index to a corresponding section object. The
        SymbolTableIndexSection was added in pyelftools 0.27.
        """
        self._shndx_sections = {}
        for x in self.elffile.iter_sections():
            if isinstance(x, et.sections.SymbolTableIndexSection):
                self._shndx_sections[x.symboltable] = x

    def get_symbols(self):
        """
        Return a set of symbols from symbol tables
        """
        symbols = {}

        # We want .symtab and .dynsym
        tables = [
            (idx, s)
            for idx, s in enumerate(self.elffile.iter_sections())
            if isinstance(s, et.sections.SymbolTableSection)
        ]

        for idx, section in tables:
            # Symbol table has no entries if this is zero
            # section.num_symbols() shows count, section.name is name
            if section["sh_entsize"] == 0:
                continue

            # We need the index of the symbol to look up versions
            for sym_idx, symbol in enumerate(section.iter_symbols()):

                # Version info is from the versym / verneed / verdef sections.
                version_info = self._get_symbol_version(section, sym_idx, symbol)

                # Symbol Type
                symbol_type = et.descriptions.describe_symbol_type(
                    symbol["st_info"]["type"]
                )

                # Symbol Binding
                binding = et.descriptions.describe_symbol_bind(
                    symbol["st_info"]["bind"]
                )

                # Symbol Visibility
                visibility = et.descriptions.describe_symbol_visibility(
                    symbol["st_other"]["visibility"]
                )

                # We aren't considering st_value, which could be many things
                # https://docs.oracle.com/cd/E19683-01/816-1386/6m7qcoblj/index.html#chapter6-35166
                symbols[symbol.name] = {
                    "version_info": version_info,
                    "type": symbol_type,
                    "binding": binding,
                    "visibility": visibility,
                    "defined": et.descriptions.describe_symbol_shndx(
                        self._get_symbol_shndx(symbol, sym_idx, idx)
                    ).strip(),
                }

        return symbols

    def _get_symbol_version(self, section, sym_idx, symbol):
        """
        Given a section, symbol index, and symbol, return version info.

        https://github.com/eliben/pyelftools/blob/master/scripts/readelf.py#L400
        """
        version_info = ""

        # I'm not sure why this would be empty
        if not self._versions:
            return version_info

        # readelf doesn't display version info for Solaris versioning
        if section["sh_type"] == "SHT_DYNSYM" and self._versions["type"] == "GNU":
            version = self._symbol_version(sym_idx)
            if version["name"] != symbol.name and version["index"] not in (
                "VER_NDX_LOCAL",
                "VER_NDX_GLOBAL",
            ):

                # This is an external symbol
                if version["filename"]:
                    version_info = "@%(name)s (%(index)i)" % version

                # This is an internal symbol
                elif version["hidden"]:
                    version_info = "@%(name)s" % version
                else:
                    version_info = "@@%(name)s" % version
        return version_info

    def _symbol_version(self, idx):
        """
        Get symbol version information.

        We can get version information for a symbol based on it's index
        https://github.com/eliben/pyelftools/blob/master/scripts/readelf.py#L942
        """
        symbol_version = dict.fromkeys(("index", "name", "filename", "hidden"))

        # No version information available
        if (
            not self._versions.get("versym")
            or idx >= self._versions.get("versym").num_symbols()
        ):
            return None

        symbol = self._versions["versym"].get_symbol(idx)
        index = symbol.entry["ndx"]
        if index not in ("VER_NDX_LOCAL", "VER_NDX_GLOBAL"):
            index = int(index)

            # GNU versioning means highest bit is used to store symbol visibility
            if self._versions["type"] == "GNU":
                if index & 0x8000:
                    index &= ~0x8000
                    symbol_version["hidden"] = True

            if (
                self._versions.get("verdef")
                and index <= self._versions["verdef"].num_versions()
            ):
                _, verdaux_iter = self._versions["verdef"].get_version(index)
                symbol_version["name"] = next(verdaux_iter).name
            else:
                verneed, vernaux = self._versions["verneed"].get_version(index)
                symbol_version["name"] = vernaux.name
                symbol_version["filename"] = verneed.name

        symbol_version["index"] = index
        return symbol_version

    def _get_symbol_shndx(self, symbol, symbol_index, symtab_index):
        """Every symbol table entry is defined in relation to some section.
        The st_shndx of a symbol holds the relevant section header table index.
        https://github.com/eliben/pyelftools/blob/master/scripts/readelf.py#L994
        """
        if symbol["st_shndx"] != et.constants.SHN_INDICES.SHN_XINDEX:
            return symbol["st_shndx"]

        # Check for or lazily construct index section mapping (symbol table
        # index -> corresponding symbol table index section object)
        if self._shndx_sections is None:
            self._shndx_sections = {}
            for sec in self.elffile.iter_sections():
                if isinstance(sec, et.sections.SymbolTableIndexSection):
                    self._shndx_sections[sec.symboltable] = sec
        return self._shndx_sections[symtab_index].get_section_index(symbol_index)

    def get_dynamic_tags(self):
        """Get the dyamic tags in the ELF file."""
        tags = {}
        for section in self.elffile.iter_sections():
            if not isinstance(section, et.dynamic.DynamicSection):
                continue

            # We are interested in architecture, soname, and needed
            def add_tag(section, tag):
                if section not in tags:
                    tags[section] = []
                tags[section].append(tag)

            for tag in section.iter_tags():
                if tag.entry.d_tag == "DT_NEEDED":
                    add_tag("needed", tag.needed)
                elif tag.entry.d_tag == "DT_RPATH":
                    add_tag("rpath", tag.rpath)
                elif tag.entry.d_tag == "DT_RUNPATH":
                    add_tag("runpath", tag.runpath)
                elif tag.entry.d_tag == "DT_SONAME":
                    tags["soname"] = tag.soname

            return tags


class Corpus(CorpusBase):
    """
    Generate an ABi corpus.

    A Corpus is an ELF file header combined with complete elf symbols,
    variables, and nested Dwarf Information Entries
    """

    def read_corpus(self):
        """
        Read the entire elf corpus, including dynamic and other sections.
        """
        reader = CorpusReader(self.path)

        # Read in the header section as part of the corpus
        self.elfheader = reader.header

        # Read in dynamic tags, and symbols
        self.dynamic_tags = reader.get_dynamic_tags()
        self.architecture = reader.get_architecture()
        self.elfclass = reader.get_elf_class()
        self.symbols = reader.get_symbols()
